------------------------------------------------------
-- 3DS Import Utility for gmax, v1.1				--
-- by Chris Cookson (cjcookson@hotmail.com)			--
--													--
-- All code (c) Copyright Chris Cookson, 2001		--
-- Please use this script any way you wish - just	--
-- drop me a line if you do something cool with it!	--
-- Oh, and if you use any/all of the code in your	--
-- own scripts, just credit me somewhere.			--
-- 													--
-- Notes:											--
--	Uses MAX4 specific features (BinStreams)		--
--													--
-- Change log:										--
--	1.1		Added support for keyframed animation.	--
--			Basic light import						--
--			Improved material import				--
--  1.0		First release							--
------------------------------------------------------

utility Import3DS "3DS Import"
(
	----------
	-- Vars --
	----------	
	local masterScale = 1.0
	local material
	local currentPath
	local fileSize=1
	local modelName = "Unnamed"
	
	local nodeNames = #()		-- Names of keyframer node objects
	
	----------------
	-- Structures --
	----------------

	struct Chunk	-- 3DS Chunk
	(
		id, name, offset, size, bstream,
		
		-- Get stream pointing to chunk data
		fn GetDataStream=
		(
			fseek bstream (offset + 6) #seek_set
			bstream
		)
	)
	
	-- Walks through chunks in the same hierarchy level
	-- To process the subchunks of a chunk, use GetSubChunkRover
	struct ChunkRover
	(
		pos, end, bstream,
		
		-- Helper function to init the top-level chunk rover
		fn Init stream=
		(
			bstream = stream
			pos = ftell bstream
			fseek bstream 0 #seek_end
			end = ftell bstream
		),
		
		-- Read the next sibling chunk
		fn ReadChunk=
		(
			if pos < end then
			(
				fseek bstream pos #seek_set
				local id = ReadShort bstream #unsigned
				local size = ReadLong bstream #unsigned
				local name = ChunkRover.GetName id
				local offset = pos
				pos += size
				Chunk id:id name:name offset:offset size:size bstream:bstream
			)
			else
			(
				undefined
			)
		),
		
		-- Get a subchunk rover for a chunk
		fn GetSubChunkRover chunk=
		(
			ChunkRover bstream:chunk.bstream pos:(chunk.offset + 6) end:(chunk.offset + chunk.size) 			
		),
		
		-- Skip a number of bytes before next chunk is read
		fn SkipBytes num=
		(
			pos += num
		),

		-- Translate a chunk ID into a semi-understandable name :)
		-- Yes, I know this can be cut down a little... later maybe
		fn GetName id=
		(
			case id of
			(
				-- Main chunks
				0x4d4d: #MAIN3DS

				0x0002: #VERSION
				0x0003: #KEYF_VERSION

				-- Colour chunks
				0x0010: #COL_FLOAT
				0x0011: #COL_24
				0x0012: #LIN_COL_24
				0x0013: #LIN_COL_FLOAT
				
				-- Info
				0x0030: #INT_PERCENTAGE		-- short percentage
				0x0031: #FLOAT_PERCENTAGE	-- float percentage
				0x0100: #MASTER_SCALE		-- float scale
				0x0995: #CHUNK_TYPE
				0x0996: #CHUNK_UNIQUE
				0x0997: #NOT_CHUNK
				0x0998: #CONTAINER
				0x0999: #ISCHUNK
				
				-- Sub chunks for EDIT3DS
				0xafff: #EDIT_MATERIAL
				0x4000: #EDIT_OBJECT
				
				-- Bitmaps
				0x1100: #BITMAP			-- cstr filename
				0x1101: #USE_BITMAP
				0x1200: #SOLID_BG		-- color_f
				0x1201: #USE_SOLID_BG
				0x1300: #V_GRADIENT		-- color_f: start, mid, end
				0x1301: #USE_V_GRADIENT
				0x1400: #LO_SHADOW_BIAS	-- float bias
				0x1410: #HI_SHADOW_BIAS
				0x1420: #SHADOW_MAP_SIZE	-- short size
				0x1430: #SHADOW_SAMPLES
				0x1440: #SHADOW_RANGE
				0x1450: #SHADOW_FILTER	-- float filter
				0x1460: #RAY_BIAS		-- float bias
				0x1500: #O_CONSTS		-- float planexyz
				
				-- Global scene options
				0x2100: #AMBIENT_LIGHT
				0x2200: #FOG				-- color_f fogbg, float nrpln, nrdens,farpln,fardens
				0x2201: #USE_FOG
				0x2210: #FOG_BG
				0x2300: #DISTANCE_CUE	-- as for FOG
				0x2301: #USE_DISTANCE_CUE
				0x2303: #USE_LAYER_FOG
				0x2310: #DCUE_BG
				0x2d2d: #SMAGIC
				0x2d3d: #LMAGIC
				
				-- Views
				0x3000: #DEFAULT_VIEW
				0x3010: #VIEW_TOP		-- float targxyz, viewwidth
				0x3020: #VIEW_BOTTOM		-- float targx, targy, targz, view_width
				0x3030: #VIEW_LEFT		-- ..
				0x3040: #VIEW_RIGHT		-- ..
				0x3050: #VIEW_FRONT		-- ..
				0x3060: #VIEW_BACK
				0x3070: #VIEW_USER
				0x3080: #VIEW_CAMERA		-- cstr cam name
				0x3090: #VIEW_WINDOW

				-- Main chunks
				0x3d3d: #EDIT3DS
				0x3d3e: #MESH_VERSION
				0x3daa: #MATLIB			-- MLI file
				0x3dc2: #PROJECT			-- 3DS Project
				0x3dff: #MATFILE			-- MAT file
		
				-- Sub chunks for EDIT_OBJECT
				0x4010: #OBJ_HIDDEN
				0x4011: #OBJ_VIS_LOFTER
				0x4012: #OBJ_DOESNT_CAST
				0x4013: #OBJ_MATTE
				0x4014: #OBJ_FAST
				0x4015: #OBJ_PROCEDURAL
				0x4016: #OBJ_FROZEN
				0x4017: #OBJ_DONT_RECEIVE_SHADOW

				0x4100: #OBJ_TRIMESH
				0x4600: #OBJ_LIGHT
				0x4700: #OBJ_CAMERA

				-- Trimesh subchunks
				0x4110: #OBJ_VERTEXLIST
				0x4111: #OBJ_VERTEXFLAGS
				0x4120: #OBJ_FACELIST
				0x4130: #OBJ_MATERIAL_GROUP	-- Material groups for faces
				0x4131: #OBJ_OLD_MATERIAL
				0x4140: #OBJ_TEXCOORDS
				0x4150: #OBJ_SMOOTHGROUP		-- Smoothing groups
				0x4160: #OBJ_LOCAL_MATRIX		-- 4x3 matrix defining local co-ord sys
				0x4165: #OBJ_MESH_COLOUR		-- short Colour index
				0x4170: #OBJ_TEXINFO			-- short maptype, xtiling etc.

				0x4181: #OBJ_PROC_NAME
				0x4182: #OBJ_PROC_DATA
				0x4190: #OBJ_MESH_BOXMAP
				
				0x4400: #OBJ_N_DL_OLD
				0x4500: #OBJ_N_CAM_OLD
				0x4600: #OBJ_N_DIRECT_LIGHT	-- colour_f
				
				-- Direct lighting chunks
				0x4610: #OBJ_DL_SPOTLIGHT	-- float targx,y,z, ang, falloffang
				0x4620: #OBJ_DL_OFF
				0x4625: #OBJ_DL_ATTENUATE
				0x4627: #OBJ_DL_RAYSHADE
				0x4630: #OBJ_DL_SHADOWED
				0x4640: #OBJ_DL_LOCAL_SHADOW
				0x4641: #OBJ_DL_LOCAL_SHADOW2
				0x4650: #OBJ_DL_SEE_CONE
				0x4651: #OBJ_DL_SPOT_RECTANGULAR
				0x4652: #OBJ_DL_SPOT_OVERSHOOT
				0x4653: #OBJ_DL_SPOT_PROJECTOR
				0x4654: #OBJ_DL_EXCLUDE
				0x4655: #OBJ_DL_RANGE
				0x4656: #OBJ_DL_SPOT_ROLL	-- float rollang
				0x4657: #OBJ_DL_SPOT_ASPECT
				0x4658: #OBJ_DL_RAY_BIAS		-- float bias
				0x4659: #OBJ_DL_INNER_RANGE	-- float range
				0x465a: #OBJ_DL_OUTER_RANGE	-- float range
				0x465b: #OBJ_DL_MULTIPLER	-- float multiple

				-- Camera subchunks
				0x4680: #OBJ_N_AMBIENT_LIGHT
				0x4700: #OBJ_N_CAMERA		-- float camxyz, targxyz, bankang, focus
				0x4710: #OBJ_CAM_SEE_CONE
				0x4720: #OBJ_CAM_RANGES		-- float nearrg, farrg

				-- Misc
				0x4d4d: #OBJ_3DSMAGIC		-- Magic no for .3DS files
				0x4f00: #OBJ_HIERARCHY
				0x4f10: #OBJ_PARENT
				0x4f20: #OBJ_PIVOT
				0x4f30: #OBJ_PIVOT_LIMITS
				0x4f40: #OBJ_PIVOT_ORDER
				0x4f50: #OBJ_XLATE_RANGE

				-- Viewports
				0x7000: #VIEWPORT_LAYOUT_OLD
				0x7001: #VIEWPORT_LAYOUT
				0x7010: #VIEWPORT_DATA_OLD
				0x7011: #VIEWPORT_DATA
				0x7012: #VIEWPORT_DATA_3
				0x7020: #VIEWPORT_SIZE
				0x7030: #NETWORK_VIEW
				
				-- Materials
				0xa000: #MAT_NAME
				0xa010: #MAT_AMBIENT			-- colour
				0xa020: #MAT_DIFFUSE			-- colour
				0xa030: #MAT_SPECULAR		-- colour
				0xa040: #MAT_SHININESS		-- % chunk
				0xa041: #MAT_SHIN2PCT		-- % chunk
				0xa042: #MAT_SHIN3PCT		-- % chunk
				0xa050: #MAT_TRANSPARENCY	-- % chunk
				0xa052: #MAT_XPFALL			-- % chunk
				0xa053: #MAT_REFBLUR			-- % chunk
				0xa080: #MAT_SELF_ILLUM
				0xa081: #MAT_TWO_SIDE
				0xa082: #MAT_DECAL
				0xa083: #MAT_ADDITIVE
				0xa084: #MAT_DECAL
				0xa085: #MAT_WIRE
				0xa086: #MAT_SUPERSAMPLE
				0xa087: #MAT_WIRESIZE		-- float wiresize
				0xa088: #MAT_FACEMAP
				0xa08a: #MAT_XPFALLIN
				0xa08c: #MAT_PHONGSOFT
				0xa08e: #MAT_WIREABS
				0xa100: #MAT_SHADING			-- short shading_value
				0xa200: #MAT_TEXMAP			-- %, mapname, maptiling, maptexblur
				0xa204: #MAT_SPECMAP			-- %, mapname
				0xa210: #MAT_OPACMAP			-- %, mapname
				0xa220: #MAT_REFLMAP			-- %, mapname
				0xa230: #MAT_BUMPMAP			-- %, mapname
				0xa240: #MAT_USE_XPFALL
				0xa250: #MAT_USE_REFBLUR
				0xa252: #MAT_BUMP_PERCENT
				0xa300: #MAT_MAP_NAME			-- cstr filename
				0xa310: #MAT_ACUBIC
				0xa320: #MAT_SXP_TEXT_DATA
				0xa321: #MAT_SXP_TEXT2_DATA
				0xa322: #MAT_SXP_BUMP_DATA
				0xa323: #MAT_SXP_SPEC_DATA
				0xa326: #MAT_SXP_SHIN_DATA
				0xa328: #MAT_SXP_SELFI_DATA
				0xa32a: #MAT_SXP_TEXT_MASKDATA
				0xa32c: #MAT_SXP_TEXT2_MASKDATA
				0xa32e: #MAT_SXP_OPAC_MASKDATA
				0xa330: #MAT_SXP_BUMP_MASKDATA
				0xa332: #MAT_SXP_SPEC_MASKDATA
				0xa334: #MAT_SXP_SHIN_MASKDATA
				0xa336: #MAT_SXP_SELFI_MASKDATA
				0xa338: #MAT_SXP_REFL_MASKDATA
				0xa33a: #MAT_TEX2MAP
				0xa33c: #MAT_SHINMAP
				0xa33d: #MAT_SELFIMAP
				0xa33e: #MAT_TEXMASK
				0xa340: #MAT_TEX2MASK
				0xa342: #MAT_OPACMASK
				0xa344: #MAT_BUMPMASK
				0xa346: #MAT_SHINMASK
				0xa348: #MAT_SPECMASK
				0xa34a: #MAT_SELFIMASK
				0xa34c: #MAT_REFLMASK
				0xa350: #MAT_MAP_TILINGOLD
				0xa351: #MAT_MAP_TILING		-- short flags
				0xa352: #MAT_MAP_TEXBLUR_OLD
				0xa353: #MAT_MAP_TEXBLUR		-- float blur
				0xa354: #MAT_MAP_USCALE
				0xa356: #MAT_MAP_VSCALE
				0xa358: #MAT_MAP_UOFFSET
				0xa35a: #MAT_MAP_VOFFSET
				0xa35c: #MAT_MAP_ANG
				0xa360: #MAT_MAP_COL1
				0xa362: #MAT_MAP_COL2
				0xa364: #MAT_MAP_RCOL
				0xa366: #MAT_MAP_GCOL
				0xa368: #MAT_MAP_BCOL
				
				-- Sub chunks for KEYF3DS
				0xb000: #KEYF3DS			-- Main Keyframer Chunk
				0xb001: #AMBIENT_NODE_TAG
				0xb002: #OBJECT_NODE_TAG	-- node_hdr, pivot, pos_track_tag, rot_track_tag, scl_track_tag, morph_smooth
				0xb003: #CAMERA_NODE_TAG	-- node_hdr, pos_track_tag, fov_track_tag, roll_track_tag
				0xb004: #TARGET_NODE_TAG	-- node_hdr, pos_track_tag
				0xb005: #LIGHT_NODE_TAG		-- node_hdr, pos_track_tag, col_track_tag
				0xb006: #L_TARGET_NODE_TAG	-- node_id, node_hdr, pos_track_tag
				0xb007: #SPOTLIGHT_NODE_TAG	-- node_id, node_hdr, pos_track_tag, hot_track_tag, fall_track_tag, roll_track_tag, col_track_tag
				0xb008: #KEYF_SEG			-- short start, end
				0xb009: #KEYF_CURTIME		-- short curFrame
				0xb00a: #KEYF_HDR			-- viewport_layout, keyf_seg, keyf_curtime, object_node_tag, light_node_tag, camera_node_tag, l_target_node_tag
											-- spotlight_node_tag, ambient_node_tag
											-- short revision, cstr filename, short animlen
				0xb010: #NODE_HDR			-- cstr objname, short flags, short flags2, short hierarchy?
				0xb011: #INSTANCE_NAME
				0xb012: #PRESCALE
				0xb013: #PIVOT
				0xb014: #BOUNDBOX
				0xb015: #MORPH_SMOOTH		-- float morph_smoothing_angle_rad
				0xb020: #POS_TRACK_TAG		--
				0xb021: #ROT_TRACK_TAG		--
				0xb022: #SCL_TRACK_TAG		--
				0xb023: #FOV_TRACK_TAG
				0xb024: #ROLL_TRACK_FLAG
				0xb025: #COL_TRACK_FLAG
				0xb026: #MORPH_TRACK_TAG
				0xb027: #HOT_TRACK_TAG
				0xb028: #FALL_TRACK_TAG
				0xb029: #HIDE_TRACK_TAG
				0xb030: #NODE_ID		-- short id
				
				default: #UNKNOWN
 			)
		)
	)
	
	struct TextureInfo (texmap, amount)
	
	-- Keyframer Track
	struct KeyFramerTrack
	(
		flags, numKeys,
		
		fn ReadTrack bstream=
		(
			flags = readShort bstream #unsigned

			-- 0x2 == REPEAT, 0x3 == LOOP
			local unknown = readShort bstream #unsigned
			unknown = readShort bstream #unsigned
			unknown = readShort bstream #unsigned
			unknown = readShort bstream #unsigned
			numKeys = readLong bstream #unsigned
			ok
		),
		
		fn SetFlags theNode bits=
		(
			local lockBits = #{}
			if bit.get flags 3 then append lockBits bits[1]
			if bit.get flags 4 then append lockBits bits[2]
			if bit.get flags 5 then append lockBits bits[3]
			local prevFlags = getTransformLockFlags theNode
			setTransformLockFlags theNode (prevFlags + lockBits)

			-- Inheritance
			lockBits = #{}
			if not bit.get flags 8 then append lockBits bits[1]
			if not bit.get flags 9 then append lockBits bits[2]
			if not bit.get flags 10 then append lockBits bits[3]
			
			prevFlags = getInheritanceFlags theNode
			setInheritanceFlags theNode (prevFlags + lockBits)
			ok
		),
		
		fn SetPositionFlags theNode=
		(
			SetFlags theNode #(1, 2, 3)
		),

		fn SetRotationFlags theNode=
		(
			SetFlags theNode #(4, 5, 6)
		),
		
		fn SetScaleFlags theNode=
		(
			SetFlags theNode #(7, 8, 9)
			ok
		)
		
	)
	
	-- Keyframer Key
	struct KeyFramerKey
	(
		-- TCB controller defaults
		frameNum, flags,
		tension = 25.0, continuity = 25.0, bias = 25.0,
		easeTo = 0.0, easeFrom = 0.0,

		-- Read a key from the supplied binstream
		fn ReadKey bstream=
		(
			frameNum = readLong bstream #unsigned
			flags = readShort bstream #unsigned		
			if (bit.get flags 1) do	tension 	= readFloat bstream
			if (bit.get flags 2) do continuity	= readFloat bstream
			if (bit.get flags 3) do bias		= readFloat bstream
			if (bit.get flags 4) do easeTo		= readFloat bstream
			if (bit.get flags 5) do easeFrom	= readFloat bstream
			ok
		)
	)
	
	--------------------
	-- User Interface --
	--------------------
	group "About"
	(
		label lab1 "3DS Importer v1.0a"
		hyperLink addy "by Chris Cookson" align:#center address:"mailto:cjcookson@hotmail.com" color:(color 0 0 170) hoverColor:(color 170 0 0)
	)

	group "Import"
	(
		checkbox keyfCheck "Import Keyframes" checked:true
		checkbox lightCheck "Import Lights" checked:true
		button importButton "Import 3DS..." toolTip:"Import 3DS File"
	)

	----------------------
	-- Helper Functions --
	----------------------

	-- Check an expression is true, throw exception if false
	fn Check expr msg=
	(
		if not expr do
		(
			throw msg
		)
		ok
	)

	--------------------------
	-- 3DS Import Functions --
	--------------------------	
	
	-- Read and decode all types of colour chunks
	fn ReadColour rover=
	(
		local ch = rover.ReadChunk()
		local bstream = ch.GetDataStream()
		local r = 0, g = 0, b = 0
		case ch.name of
		(
			#COL_FLOAT:
			(
				r = 255 * ReadFloat bstream
				g = 255 * ReadFloat bstream
				b = 255 * ReadFloat bstream
			)
			
			#COL_24:
			(
				r = ReadByte bstream #unsigned
				g = ReadByte bstream #unsigned
				b = ReadByte bstream #unsigned
			)
			
			#LIN_COL_24:
			(
				r = ReadByte bstream #unsigned
				g = ReadByte bstream #unsigned
				b = ReadByte bstream #unsigned
			)
			
			#LIN_COL_FLOAT:
			(
				r = 255 * ReadFloat bstream
				g = 255 * ReadFloat bstream
				b = 255 * ReadFloat bstream
			)

			default:
				format "Error: Excepted a colour chunk, but found %\n" ch.name
		)
		color r g b
	)
	
	-- Read and decode percentage chunks
	fn ReadPercentage rover=
	(
		local ch = rover.ReadChunk()
		local bstream = ch.GetDataStream()
		local percent = 0
		case ch.name of
		(
			#INT_PERCENTAGE:
				percent = ReadShort bstream #unsigned
			
			#FLOAT_PERCENTAGE:
				percent = ReadFloat bstream

			default:
				format "Error: Excepted a percentage chunk, but found %\n" ch.name
		)
		percent
	)

	------------------------
	-- Read a texture map --
	------------------------
	fn ReadTextureMap texMapChunk=
	(
		local rover = ChunkRover.GetSubChunkRover texMapChunk
		local amount = ReadPercentage rover
		
		local texChunk = rover.ReadChunk()
		
		local texMap = undefined

		while texChunk != undefined do
		(
			case texChunk.name of
			(
				#MAT_MAP_NAME:
				(
					local bmpFilename = ReadString (texChunk.GetDataStream())
					
					texMap = sceneMaterials[bmpFilename]
					if texMap != undefined then
					(
						if (classof texMap) != BitmapTexture then
						(
							texMap = undefined
						) else
						(
							return TexInfo amount:amount texmap:texMap
						)
					)
					
					try
					(
						local bmp = openBitMap (currentPath + bmpFilename)
						texMap = BitmapTexture bitmap:bmp name:bmpFileName
					) catch
					(
						format "Warning: Can't find texture map file %\n" bmpFilename
						return undefined
					)
				)
				
				#MAT_MAP_TEXBLUR:
					texMap.coords.blur = ReadFloat (texChunk.GetDataStream())
				
				#MAT_MAP_USCALE:
					texMap.coords.U_Tiling = ReadFloat (texChunk.GetDataStream())
				
				#MAT_MAP_VSCALE:
					texMap.coords.V_Tiling = ReadFloat (texChunk.GetDataStream())
				
				#MAT_MAP_UOFFSET:
					texMap.coords.U_Offset = ReadFloat (texChunk.GetDataStream())
				
				#MAT_MAP_VOFFSET:
					texMap.coords.V_Offset = ReadFloat (texChunk.GetDataStream())
					
				default:
				(
					--format "Found texmap chunk %\n" texChunk.name
				)
			)
			texChunk = rover.ReadChunk()
		)
		
		TextureInfo amount:amount texmap:texMap
	)

	---------------------------------
	-- Read an EDIT_MATERIAL chunk --
	---------------------------------
	fn ReadMaterial matTopChunk=
	(
		local rover = ChunkRover.GetSubChunkRover matTopChunk
		local matChunk = rover.ReadChunk()
		
		-- Create the material
		local mat = standard name:"My new material"
		mat.shaderType = 1	-- Blinn
		
		local matName = "Unnamed Material"
		
		while matChunk != undefined do
		(
			local subRover = ChunkRover.GetSubChunkRover(matChunk)
			
			case matChunk.name of
			(
				#MAT_NAME:
				(
					matName = ReadString (matChunk.GetDataStream())				
					mat.name = matName
				)

				#MAT_AMBIENT:
					mat.ambient = ReadColour subRover

				#MAT_DIFFUSE:
					mat.diffuse = ReadColour subRover

				#MAT_SPECULAR:
					mat.specular = ReadColour subRover

				#MAT_SHININESS: -- ratio
					mat.specularLevel = ReadPercentage subRover
					
				#MAT_SHIN2PCT: -- strength?
					mat.glossiness = ReadPercentage subRover
				
--				#MAT_XPFALL: -- transparency falloff
				
				#MAT_SHADING: -- shading style
				(
					local shadType = readShort (matChunk.GetDataStream()) #unsigned
					case shadType of
					(
						-- 0 is wireframe
						1: mat.faceted = true -- flat
						2: mat.shaderType = 1 -- gouraud
						3: mat.shaderType = 5 -- phong
						4: mat.shaderType = 2 -- metal
					)
				)
				
				#MAT_SELF_ILLUM: -- self-illum
				(
					mat.useSelfIllumColor = true
					mat.selfIllumColor = ReadColour subRover
				)

				#MAT_SELF_ILPCT: -- self-illum strength
					mat.selfIllumAmount = ReadPercentage subRover
				
				#MAT_TWO_SIDE:
					mat.twoSided = true
				
				#MAT_FACEMAP:
					mat.faceMap = true

--				#MAT_TRANSPARENCY:
--					mat.opacity = ReadPercentage subRover

				#MAT_TEXMAP:
				(
					local tmapInfo = ReadTextureMap matChunk

					if tmapInfo != undefined then
					(
						mat.mapAmounts[2] = tmapInfo.amount
						mat.maps[2] = tmapInfo.texmap
						mat.mapEnables[2] = true
						showTextureMap mat tmapInfo.texmap true
					)
				)
				
				#MAT_BUMPMAP:
				(
					local tmapInfo = ReadTextureMap matChunk
					if tmapInfo != undefined then
					(
						mat.mapAmounts[9] = tmapInfo.amount
						mat.maps[9] = tmapInfo.texmap
						mat.mapEnables[9] = true
						showTextureMap mat tmapInfo.texmap true
					)
				)
				
				#MAT_OPACMAP:
				(
					local tmapInfo = ReadTextureMap matChunk
					if tmapInfo != undefined then
					(
						mat.mapAmounts[7] = tmapInfo.amount
						mat.maps[7] = tmapInfo.texmap
						mat.mapEnables[7] = true
						showTextureMap mat tmapInfo.texmap true
					)
				)
				
--				default:
--					format "Found %\n" matChunk.name
			)
			matChunk = rover.ReadChunk()
		)
		
		-- Add material to our big fat multimaterial
		material.numsubs += 1
		local idx = material.numsubs
		material[idx] = mat
		material.names[idx] = matName
		ok
	)

	----------------------------------
	-- Read an N_DIRECT_LIGHT chunk --
	----------------------------------
	fn ReadDirectLight lightName lightChunk=
	(
		local rover = ChunkRover.GetSubChunkRover lightChunk
		
		-- Get light position in world coordsys
		local bstream = lightChunk.GetDataStream()
		local lightPos = [readFloat bstream, readFloat bstream, readFloat bstream]
		rover.SkipBytes 12
		
		-- Create a simple omni
		local newLight = omniLight name:lightName
		
		-- Read rest of light chunks
		local curChunk = rover.ReadChunk()
		while curChunk != undefined do
		(
			bstream = curChunk.GetDataStream()
			
			case curChunk.name of
			(
				#COL_FLOAT:
					newLight.rgb = color (readFloat bstream) (readFloat bstream) (readFloat bstream)
				
				#DL_OFF:
					newLight.enabled = false
				
				#DL_ATTENUATE:
					newLight.useFarAtten = true

				#DL_OUTER_RANGE:
					newLight.farAttenEnd = readFloat bstream
				
				#DL_INNER_RANGE:
					newLight.farAttenStart = readFloat bstream
				
				#DL_MULTIPLIER:
					newLight.multiplier = readFloat bstream
				
				#DL_EXCLUDE:
				(
					-- Exclude this object
					local excludedName = readString bstream
					append newLight.excludeList (for obj in objects where obj.name == excludedName collect obj)
				)
				
				#DL_SPOTLIGHT:
				(
					-- ... lots of spotlight stuff, will add this later
				)
			)
			
			curChunk = rover.ReadChunk()
		)
		ok
	)


	-------------------------------
	-- Read an EDIT_OBJECT chunk --
	-------------------------------
	fn ReadEditObject editObjChunk=
	(
	
		-- First item is ASCIIZ object name
		local objName = ReadString (editObjChunk.GetDataStream())
		
		-- Pull out the first object chunk
		local rover = ChunkRover.GetSubChunkRover editObjChunk
		rover.SkipBytes( objName.count + 1 )	-- skip object name bytes
		local objChunk = rover.ReadChunk()
		
		while objChunk != undefined do
		(
			-- Object could be camera, light or mesh
			if objChunk.name == #OBJ_TRIMESH then
			(
				-- Keep reading different chunks collecting data
				local meshRover = ChunkRover.GetSubChunkRover objChunk

				local verts = #()
				local faces = #()
				local texcoords = #()
				local smoothGroups = #()
				local faceMaterials = #()
				
				local localMatrix = matrix3 1

				local curChunk = meshRover.ReadChunk()
				while curChunk != undefined do
				(
					local bstream = curChunk.GetDataStream()
					case curChunk.name of
					(
						#OBJ_VERTEXLIST:
						(
							-- Get number of verts and read 'em in
							local numVerts = ReadShort bstream #unsigned

							for i = 1 to numVerts do
							(
								local x = ReadFloat bstream
								local y = ReadFloat bstream
								local z = ReadFloat bstream

								append verts [x, y, z]
							)
						)

						-- Face list
						#OBJ_FACELIST:
						(
							local numFaces = ReadShort bstream #unsigned
							for i = 1 to numFaces do
							(
								local v1 = ReadShort bstream #unsigned
								local v2 = ReadShort bstream #unsigned
								local v3 = ReadShort bstream #unsigned
								local faceInfo = ReadShort bstream #unsigned

								-- Note: 3DS indices are 0 based
								append faces [v1+1, v2+1, v3+1]
							)
							
							local faceRover = ChunkRover.GetSubChunkRover curChunk
							faceRover.SkipBytes (2+(numFaces * 8))
							
							local faceChunk = faceRover.ReadChunk()
							while faceChunk != undefined do
							(
								if faceChunk.name == #OBJ_MATERIAL_GROUP do
								(
									local materialName = ReadString bstream
		
									local numFaces = ReadShort bstream #unsigned
									for i = 1 to numFaces do
									(
										local faceIndex = ReadShort bstream #unsigned
										faceIndex += 1
										faceMaterials[faceIndex] = materialName
									)
								)
								if faceChunk.name == #OBJ_SMOOTHGROUP do
								(
									local str = faceChunk.GetDataStream()
									
									for i = 1 to faces.count do
									(
										local faceGroups = readLong str #unsigned
										append smoothGroups faceGroups
									)
								)
								faceChunk = faceRover.ReadChunk()
							)
							
						)

						#OBJ_TEXCOORDS:
						(
							local numVerts = ReadShort bstream #unsigned
							for i = 1 to numVerts do
							(
								local u = ReadFloat bstream
								local v = ReadFloat bstream
								append texCoords [u, v, 0.0]
							)
						)
						
						#OBJ_LOCAL_MATRIX:
						(
							local rx = [(ReadFloat bstream), (ReadFloat bstream), (ReadFloat bstream)]
							local ry = [(ReadFloat bstream), (ReadFloat bstream), (ReadFloat bstream)]
							local rz = [(ReadFloat bstream), (ReadFloat bstream), (ReadFloat bstream)]
							local rt = [(ReadFloat bstream), (ReadFloat bstream), (ReadFloat bstream)]
							
							localMatrix = matrix3 rx ry rz rt
						)
						
						#OBJ_MESH_COLOUR:
						(
						)

						-- Ignore all other chunks (for now, anyway)
						default:
						(
							format "EDIT_OBJ: Found chunk %\n" curChunk.name
						)
					)
					curChunk = meshRover.ReadChunk()
				)

				-- Make sure we have at least 1 standard material available
				if material.numsubs == 0 do
				(
					material.numsubs = 1
					material[1] = standard name:"Substituted Default material"
				)
				
				if (faces.count > 0) and (verts.count > 0) then
				(
					local s = [masterScale, masterScale, masterScale]
					
					-- Make an editable mesh
					local newMesh = mesh name:objName position:[0, 0, 0] scale:s \
									faces:faces vertices:verts material:material

					-- Transform all verts into object space
					in coordsys world
					(
						-- localMatrix is object->world
						newMesh.transform = localMatrix
				
						-- world->object space
						local worldToObject = inverse localMatrix

						newMesh.objectoffsetpos = worldToObject.translationPart
						newMesh.objectoffsetrot = worldToObject.rotationPart
						newMesh.objectoffsetscale = worldToObject.scalePart
					)
							  
					-- Add texcoords, if we have them
					if texCoords.count > 0 do
					(
						setNumTVerts newMesh texCoords.count
					
						for i = 1 to texCoords.count do
						(
							setTVert newMesh i texCoords[i]
						)

						buildTVFaces newMesh
						
						for i = 1 to faces.count do
						(
							setTVFace newMesh i faces[i]
						)
					)
					
					-- Add smooth modifier for faces with no smooth group
					-- FIXME: Doesn't quite work as I thought it might :(
					addModifier newMesh (smooth autosmooth:true threshold:45)
					collapseStack newMesh

					-- Assign smoothing groups
					for i = 1 to smoothGroups.count do
					(
						setFaceSmoothGroup newMesh i smoothGroups[i]
					)
					
					-- Assign material ID's NB. materials must be already defined
					for i = 1 to faces.count do
					(
						local matID = 1
						
						for matSlot = 1 to material.numsubs \
							where material.names[matSlot] == faceMaterials[i] do
						(
							matID = material.materialIDList[matSlot]
						)
							
						setFaceMatID newMesh i matID
					)
					
					-- Update the mesh
					update newMesh
					
					-- Admire your newly imported mesh :)
				) else
				(
					print "Warning: Tri mesh is incomplete!"
				)
			)
			
			-- Import direct light
			if lightCheck.checked and (objChunk.name == #N_DIRECT_LIGHT) do
			( 
				ReadDirectLight objName objChunk 
			)
		
			-- Read next EDIT_OBJECT chunk
			objChunk = rover.ReadChunk()
		)
		ok
	)

	---------------------------------
	-- Read the main EDIT3DS chunk --
	---------------------------------
	fn ReadEdit3DS edit3dsChunk=
	(
		-- Set some global info
		masterScale = 1.0
		material = multiMaterial numsubs:0 name:modelName
		
		local rover = ChunkRover.GetSubChunkRover edit3dsChunk

		local curChunk = rover.ReadChunk()
		
		while curChunk != undefined do
		(
			case curChunk.name of
			(
				#EDIT_OBJECT:
				(
					ReadEditObject curChunk
				)

				#MASTER_SCALE:
				(
					masterScale = ReadFloat (curChunk.GetDataStream())
				)

				#EDIT_MATERIAL:
				(
					ReadMaterial curChunk
				)
			
				default:
				(
					format "Ignoring chunk '%'\n" curChunk.name
				)
			)
			curChunk = rover.ReadChunk()
			progressUpdate (100 * rover.pos / fileSize)
		)
		ok	
	)

	--------------------------------
	-- Read OBJECT_NODE_TAG chunk --
	--------------------------------
	fn ReadObjectNodeTag objectNodeTagChunk=
	(
		local rover = ChunkRover.GetSubChunkRover objectNodeTagChunk
		
		local curChunk = rover.ReadChunk()
		
		local pivotPos = [0,0,0]
		local curObject = undefined
		local curNodeID = 0
				
		while curChunk != undefined do
		(
			case curChunk.name of
			(
				#NODE_ID:
				(
					local bstream = curChunk.GetDataStream()
					curNodeID = readShort bstream #unsigned
					--format "Node ID = %\n" curNodeID
				)
				
				#NODE_HDR:
				(
					local bstream = curChunk.GetDataStream()
					local curObjectName = readString bstream
					local unk1 = readShort bstream #unsigned
					local unk2 = readShort bstream #unsigned
					local hierarchy = readShort bstream #unsigned

					--format "ObjName = %, hierarchy = %\n" curObjectName hierarchy
					--format "curNodeID+1= % curObjectName=%\n" (curNodeID+1) curObjectName

					hierarchy += 1
					
					if curObjectName == "$$$DUMMY" then
					(
						-- Create special dummy object
						local dummyObject = dummy name:(uniqueName "Dummy")
						curObjectName = dummyObject.name
						curObject = dummyObject
					)
					
					nodeNames[curNodeID + 1] = curObjectName

					local curObjectSet = for o in objects where o.name == curObjectName collect o
					Check (curObjectSet.count > 0) "NODE_HDR: Cannot find object!"
					curObject = curObjectSet[1]
					
					if hierarchy != 65536 do
					(
						local itsParent = for p in objects where p.name == nodeNames[hierarchy] collect p
						Check (itsParent.count > 0) "NODE_HDR: Cannot find parent!"
						
						curObject.parent = itsParent[1]
					)
				)
				
				#PIVOT:
				(
					local bstream = curChunk.GetDataStream()
					local x = readFloat bstream
					local y = readFloat bstream
					local z = readFloat bstream
					pivotPos = [x, y, z]
					
					curObject.objectoffsetpos -= pivotpos
				)
				
				#INSTANCE_NAME:
				(
					local bstream = curChunk.GetDataStream()
					local instName = readString bstream

					curObject.name = instName
					nodeNames[curNodeID + 1] = curObject.name
				)
				
				#POS_TRACK_TAG:
				(
					local bstream = curChunk.GetDataStream()
					
					local posTrack = KeyFramerTrack()
					posTrack.ReadTrack bstream
					
					posTrack.SetPositionFlags curObject
					
					for i = 1 to posTrack.numKeys do
					(
						local posKey = KeyFramerKey()
						posKey.ReadKey bstream
						
						local x = readFloat bstream
						local y = readFloat bstream
						local z = readFloat bstream
						
						in coordsys parent
						(
							local newKey = addNewKey curObject.pos.track posKey.frameNum
							newKey.value = [x, y, z]
						)
					)
				)
				
				#ROT_TRACK_TAG:
				(
					local bstream = curChunk.GetDataStream()
					
					local rotTrack = KeyFramerTrack()
					rotTrack.ReadTrack bstream

					rotTrack.SetRotationFlags curObject
					
					local accumQuat = quat 0 0 0 1
					for i = 1 to rotTrack.numKeys do
					(
						local curKey = KeyFramerKey()
						curKey.ReadKey bstream

						-- Rotation is angle and an axis
						local angleRad = readFloat bstream
						local x = readFloat bstream
						local y = readFloat bstream
						local z = readFloat bstream
						
						local q = quat (RadToDeg -angleRad) [x, y, z]

						in coordsys parent
						(
							local newKey = addNewKey curObject.rotation.track curKey.frameNum
							newKey.value = q
							newKey.tension = curKey.tension
							newKey.continuity = curKey.continuity
							newKey.bias = curKey.bias
							newKey.easeTo = curKey.easeTo
							newKey.easeFrom = curKey.easeFrom
						)
					)
				)
				
				#SCL_TRACK_TAG:
				(
					local bstream = curChunk.GetDataStream()
					
					local sclTrack = KeyFramerTrack()
					sclTrack.ReadTrack bstream
					
					sclTrack.SetScaleFlags curObject
					
					for i = 1 to sclTrack.numKeys do
					(
						local sclKey = KeyFramerKey()
						sclKey.ReadKey bstream
			
						local scaleX = readFloat bstream
						local scaleY = readFloat bstream
						local scalez = readFloat bstream
						
						in coordsys parent
						(
							local newKey = addNewKey curObject.scale.track sclKey.frameNum
							newKey.value = [scaleX, scaleY, scaleZ]
						)
					)
				)
				
				#BOUNDBOX:
				(
					local bstream = curChunk.GetDataStream()
					
					local bbMin = [ readFloat bstream, readFloat bstream , readFloat bstream ]
					local bbMax = [ readFloat bstream, readFloat bstream , readFloat bstream ]
					
					if classOf curObject == dummy then
					(
						curObject.boxSize = bbMax - bbMin
					)
					
				)
			)
		
			curChunk = rover.ReadChunk()		
		)
		ok	
	)


	----------------------------
	-- Read the KEYF3DS chunk --
	----------------------------
	fn ReadKeyF3DS keyfChunk=
	(
	
		local rover = ChunkRover.GetSubChunkRover keyfChunk
		
		local curChunk = rover.ReadChunk()
		
		local segStart = 0
		local segEnd = 0
		local curFrame = 0
		
		-- Reset globals
		nodeNames = #()
		
		while curChunk != undefined do
		(
--			format "KEYF: Found chunk %\n" curChunk.name
			
			case curChunk.name of
			(
				#KEYF_HDR:
				(
					local bstream = curChunk.GetDataStream()
					local revision = readShort bstream #unsigned
					-- 13 bytes (12 byte name + NULL)
					-- long - animation length in frames
				)
				
				#KEYF_SEG:
				(
					local bstream = curChunk.GetDataStream()
					
					segStart = readLong bstream #unsigned
					segEnd = readLong bstream #unsigned
			
					if segStart != segEnd then
					(
						animationRange = interval segStart segEnd
					)
				)
				
				#KEYF_CURTIME:
				(
					curFrame = readShort (curChunk.GetDataStream()) #unsigned
				)
				
				#OBJECT_NODE_TAG:
				(
					ReadObjectNodeTag curChunk
				)
			)
			
			curChunk = rover.ReadChunk()
			progressUpdate (100 * rover.pos / fileSize)
		)
		ok
	)
	

	--------------------------
	-- Main Import Function --
	--------------------------

	on importButton pressed do
	(
		-- Show open file dialog box
		local objFileName = getOpenFileName caption:"Import 3DS" types:"Autodesk 3DS (*.3ds)|*.3ds|All Files (*.*)|*.*|"
		
		-- If user made a selection, begin importing
		if objFileName != undefined then
		(
			currentPath = getFilenamePath objFileName
			modelName = GetFilenameFile objFileName
	
			-- Open up the file as a binary stream
			local bstream = fopen objFileName "rb"
								
			local isValid = true
			
			(
				local topRover = ChunkRover()
				topRover.Init bstream
				
				fileSize = topRover.end
				
				-- Show progress bar
				progressStart ("Importing " + (filenameFromPath objFileName) + "...")
				
				-- Read main 3DS chunk
				local mainChunk = topRover.ReadChunk()
				Check (mainChunk.name == #MAIN3DS) "Main 3DS chunk is missing!"
				
				-- Process subchunks
				local rover = ChunkRover.GetSubChunkRover mainChunk
				
				local curChunk = rover.ReadChunk()
				
				while curChunk != undefined do
				(
					case curChunk.name of
					(
						#EDIT3DS:
						(
							ReadEdit3DS curChunk
						)
							
						#KEYF3DS:
						(
							if keyfCheck.checked then ReadKeyF3DS curChunk
						)
						
						#VERSION:
						(
							local ver = ReadShort (curChunk.GetDataStream()) #unsigned
							format "3DS Version: %\n" ver
						)
						
						default:
							format "Warning, found unknown main chunk! %\n" curChunk
					)
					curChunk = rover.ReadChunk()
				)
				progressEnd()
				
			)
			
			fclose bstream
			
			max views redraw
		)
	)
	
)

